package net.lucamasini.testcase.server.boundary;

import com.google.common.collect.Multimap;
import lombok.extern.slf4j.Slf4j;
import net.lucamasini.testcase.server.boundary.resource.HasFindByFilters;
import net.lucamasini.testcase.server.control.ContatoreComandaControl;
import net.lucamasini.testcase.server.control.OrdineControl;
import net.lucamasini.testcase.server.producers.BarUtente;
import net.lucamasini.testcase.shared.dto.Canale;
import net.lucamasini.testcase.shared.dto.PaginatedSearchResultWithLastModified;
import net.lucamasini.testcase.shared.dto.StatoOrdine;
import net.lucamasini.testcase.shared.entity.Bar;
import net.lucamasini.testcase.shared.entity.BarLayout;
import net.lucamasini.testcase.shared.entity.Ordine;
import net.lucamasini.testcase.shared.entity.PiattoOrdinato;
import org.eclipse.microprofile.config.inject.ConfigProperty;
import org.jboss.resteasy.annotations.cache.NoCache;

import javax.enterprise.context.RequestScoped;
import javax.inject.Inject;
import javax.persistence.EntityManager;
import javax.persistence.TemporalType;
import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.Order;
import javax.persistence.criteria.Predicate;
import javax.persistence.criteria.Root;
import javax.servlet.http.HttpServletRequest;
import javax.transaction.Transactional;
import javax.validation.Valid;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Pattern;
import javax.ws.rs.*;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.UriInfo;
import java.sql.Timestamp;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.util.*;

import static net.lucamasini.testcase.shared.common.SharedConstants.*;
import static net.lucamasini.testcase.shared.dto.Canale.APP;
import static net.lucamasini.testcase.shared.dto.Canale.CHIOSCO;
import static net.lucamasini.testcase.shared.dto.StatoOrdine.CHIUSO;
import static net.lucamasini.testcase.shared.dto.StatoOrdine.SOSPESO;

@Path("/ordini")
@Slf4j
@RequestScoped
public class OrdiniResource implements HasFindByFilters<Ordine> {

    @Inject
    EntityManager em;

    @Inject
    @BarUtente
    Bar barUtente;

    @ConfigProperty(name = "ordbar.mediaTypeZD410")
    @Inject
    String mediaTypeZD410;

    @Inject
    OrdineControl ordineControl;

    @Inject
    OrdiniService ordiniService;

    @Inject
    ContatoreComandaControl contatoreComandaControl;

    @Inject
    protected HttpServletRequest request;

    @Inject
    BarLayoutService barLayoutService;

    @Override
    public EntityManager getEm() {
        return em;
    }

    @Override
    public Class<Ordine> getEntityType() {
        return Ordine.class;
    }

    @GET
    @NoCache
    @Path("{id}")
    @Produces(MediaType.APPLICATION_JSON)
    public Ordine findOrdineById(@PathParam("id") @NotNull String id) {
        return ordiniService.findByTypeId(id);
    }

    @GET
    @NoCache
    @Produces(MediaType.APPLICATION_JSON)
    public Response findByFiltersOrdini(@Context UriInfo uriInfo,
                                        @HeaderParam(PAGING_FILTER_START) Integer start,
                                        @HeaderParam(PAGING_FILTER_LIMIT) Integer limit) {
        return HasFindByFilters.super.findByFilters(uriInfo, start, limit);
    }

    @GET
    @Path("/fascia/{inizioFascia}")
    @Produces(MediaType.APPLICATION_JSON)
    public List<Ordine> getOrdini(@PathParam("inizioFascia") @Pattern(regexp = "[0-2]?[0-9]:[0-9]{2}", message = "Formato orario valido hh:mm") String fascia) {

        String[] fasciaValue = fascia.split(":");

        Calendar calendar = Calendar.getInstance();
        calendar.set(Calendar.HOUR_OF_DAY, Integer.parseInt(fasciaValue[0]));
        calendar.set(Calendar.MINUTE, Integer.parseInt(fasciaValue[1]));
        calendar.set(Calendar.SECOND, 0);
        calendar.set(Calendar.MILLISECOND, 0);
        List<Ordine> resultList = this.em.createNamedQuery("Ordine.findOrdiniAppByInizioFascia", Ordine.class)
                .setParameter("today", Timestamp.valueOf(LocalDateTime.of(LocalDate.now(), LocalTime.MIDNIGHT)))
                .setParameter("idBarValue", barUtente.getId())
                .setParameter("fasciaValue", calendar.getTime(), TemporalType.TIMESTAMP)
                .getResultList();

        resultList.forEach(ordine -> ordine.getPiattiOrdinati().forEach(PiattoOrdinato::hasAbbinamenti));

        return resultList;
    }


    @GET
    @Path("stato/{stato}")
    @Produces(MediaType.APPLICATION_JSON)
    public PaginatedSearchResultWithLastModified<Ordine> getOrdiniByStato(@PathParam("stato") StatoOrdine stato) {

        List<Ordine> resultList = this.em.createNamedQuery("Ordine.findOrdiniByCanaleAndStato", Ordine.class)
                .setParameter("today", Timestamp.valueOf(LocalDateTime.of(LocalDate.now(), LocalTime.MIDNIGHT)))
                .setParameter("idBarValue", barUtente.getId())
                .setParameter("canali", Arrays.asList(Canale.values()))
                .setParameter("statoValue", stato)
                .getResultList();

        resultList.forEach(ordine -> ordine.getPiattiOrdinati().forEach(PiattoOrdinato::hasAbbinamenti));
        return new PaginatedSearchResultWithLastModified<>(resultList.size(), resultList, null);
    }


    @GET
    @Path("canale/{canale}/stato/{stato}")
    @Produces(MediaType.APPLICATION_JSON)
    public List<Ordine> getOrdini(@PathParam("canale") Canale canale,
                                  @PathParam("stato") StatoOrdine stato) {
        List<Ordine> resultList = this.em.createNamedQuery("Ordine.findOrdiniByCanaleAndStato", Ordine.class)
                .setParameter("today", Timestamp.valueOf(LocalDateTime.of(LocalDate.now(), LocalTime.MIDNIGHT)))
                .setParameter("idBarValue", barUtente.getId())
                .setParameter("canali", canale)
                .setParameter("statoValue", stato)
                .getResultList();

        resultList.forEach(ordine -> ordine.getPiattiOrdinati().forEach(PiattoOrdinato::hasAbbinamenti));
        return resultList;
    }

    @POST
    @Consumes(MediaType.APPLICATION_JSON)
    @Produces(MediaType.APPLICATION_JSON)
    @Transactional
    public Ordine saveOrUpdate(@Valid Ordine ordine) {

        if (ordine.getBarcode() == null) {
            if (ordine.getCanale() != APP) {
                ordine.setBar(getEm().find(Bar.class, barUtente.getId()));
            }

            Long numeroComanda = contatoreComandaControl.getNumeroComanda(barUtente.getId());
            ordine.setOldNumeroComanda(numeroComanda);
            ordine.setNumeroComanda(ordine.getCanale().getComandaPrefix() + numeroComanda);
            ordine.setBarcode(ordineControl.generaOrdineId(numeroComanda));
        }

        Ordine ordineMerged = ordiniService.saveOrUpdateInternal(ordine);
        if (ordine.getCanale() == CHIOSCO && ordine.getStato() == SOSPESO) {
            cambiaStato(ordineMerged, CHIUSO);
        }

        return ordineMerged;
    }

    public Response cambiaStato(Ordine ordine, StatoOrdine statoTarget) {
        BarLayout barLayout = barLayoutService.getBarLayout(ordine.getBar().getId());
        StatoOrdine.OrdineContestoStato context = new StatoOrdine.OrdineContestoStato(ordine, barLayout);

        boolean statoModificato = statoTarget.cambiaStato(context);

        Response response = Response.ok(ordine).build();
        if (!statoModificato) {
            response = Response.notModified().build();
        }
        return response;
    }


    @Override
    public List<Predicate> buildPredicates(Multimap<String, String> filters, Root<Ordine> root, Date ifModifiedSince) {
        CriteriaBuilder builder = em.getCriteriaBuilder();
        List<Predicate> predicates = new ArrayList<>();

        predicates.add(builder.greaterThanOrEqualTo(root.get("dataInserimento"), Timestamp.valueOf(LocalDateTime.of(LocalDate.now(), LocalTime.MIDNIGHT))));

        if (filters.get(FILTER_CANALE) != null && !filters.get(FILTER_CANALE).isEmpty()) {
            predicates.add(root.get(FILTER_CANALE).in(Canale.getCanali(filters.get(FILTER_CANALE))));
        }

        if (filters.get(FILTER_STATO_ORDINE) != null && !filters.get(FILTER_STATO_ORDINE).isEmpty()) {
            predicates.add(root.get("stato").in(StatoOrdine.stringsToStati(filters.get(FILTER_STATO_ORDINE))));
        }

        if (filters.get(FILTER_COMANDA) != null && !filters.get(FILTER_COMANDA).isEmpty()) {
            predicates.add(root.get("numeroComanda").in(filters.get(FILTER_COMANDA)));
        }

        if (filters.get(FILTER_BAR) != null && !filters.get(FILTER_BAR).isEmpty()) {
            predicates.add(root.get("bar").get("id").in(filters.get(FILTER_BAR)));
        }

        if (ifModifiedSince != null) {
            predicates.add(builder.and(
                    builder.isNotNull(root.get("dataAggiornamento")),
                    builder.greaterThan(root.<Timestamp>get("dataAggiornamento"), ifModifiedSince)));
        }

        return predicates;
    }

    @Override
    public void addJoins(Root<Ordine> root) {

    }

    @Override
    public List<Order> orderBy(Root<Ordine> root) {
        CriteriaBuilder builder = em.getCriteriaBuilder();
        List<Order> orders = new ArrayList<>();
        orders.add(builder.asc(root.get("dataInserimento")));
        return orders;
    }
}
