package net.lucamasini.testcase.shared.dto;

import lombok.AllArgsConstructor;
import lombok.Getter;
import lombok.NoArgsConstructor;
import net.lucamasini.testcase.shared.entity.BarLayout;
import net.lucamasini.testcase.shared.entity.FlussoServizio;
import net.lucamasini.testcase.shared.entity.Ordine;
import net.lucamasini.testcase.shared.entity.Piatto;
import org.jboss.errai.common.client.api.annotations.Portable;

import java.time.Instant;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.*;

@Getter
@Portable
public enum StatoOrdine implements HasTransition<StatoOrdine> {

    APERTO {
        @Override
        public List<StatoOrdine> getProssimiStati() {
            return Arrays.asList(StatoOrdine.SOSPESO, StatoOrdine.CHIUSO, StatoOrdine.CANCELLATO, StatoOrdine.SCARTATO);
        }
    },
    SOSPESO {
        @Override
        public List<StatoOrdine> getProssimiStati() {
            return Collections.singletonList(StatoOrdine.CHIUSO);
        }

        @Override
        public void postCambiaStato(StatoOrdine.OrdineContestoStato context) {
            context.actions().sospeso(context);
        }
    },
    CHIUSO {
        @Override
        public List<StatoOrdine> getProssimiStati() {
            return Arrays.asList(StatoOrdine.EVASO, StatoOrdine.PIATTI_IN_PREPARAZIONE, StatoOrdine.BEVANDE_IN_PREPARAZIONE, StatoOrdine.CANCELLATO, StatoOrdine.NON_RITIRATO);
        }

        @Override
        public void postCambiaStato(StatoOrdine.OrdineContestoStato context) {
            context.actions().chiuso(context);
            context.publish(Device.MONITOR_VASSOIO, Device.POSTAZIONE_FRONT);
        }
    },
    RECUPERATO {
        @Override
        public List<StatoOrdine> getProssimiStati() {
            return Arrays.asList(StatoOrdine.EVASO, StatoOrdine.PIATTI_IN_PREPARAZIONE, StatoOrdine.BEVANDE_IN_PREPARAZIONE);
        }

        @Override
        public void postCambiaStato(StatoOrdine.OrdineContestoStato context) {
            context.publish(Device.MONITOR_VASSOIO);
        }
    },
    CANCELLATO {
        @Override
        public List<StatoOrdine> getProssimiStati() {
            return Collections.singletonList(StatoOrdine.SCARTATO);
        }

        @Override
        public void preCambiaStato(StatoOrdine.OrdineContestoStato context) {
            if (!context.actions().isCancellazioneAllowed(context)) {
                throw new RuntimeException("for test purpose");
            }
        }

        @Override
        public void postCambiaStato(StatoOrdine.OrdineContestoStato context) {
            context.actions().cancellato(context);
            context.publish(Device.MONITOR_VASSOIO);
        }
    },
    PIATTI_IN_PREPARAZIONE {
        @Override
        public List<StatoOrdine> getProssimiStati() {
            return Arrays.asList(StatoOrdine.BEVANDE_IN_PREPARAZIONE, StatoOrdine.CANCELLATO);
        }

        @Override
        public void postCambiaStato(StatoOrdine.OrdineContestoStato context) {
            context.actions().piattiInPreparazione(context);
            context.publish(Device.MONITOR_VASSOIO, Device.POSTAZIONE_BACK, Device.NUMERI);
        }
    },
    BEVANDE_IN_PREPARAZIONE {
        @Override
        public List<StatoOrdine> getProssimiStati() {
            return Arrays.asList(StatoOrdine.PRONTO, StatoOrdine.IN_CONSEGNA, StatoOrdine.CANCELLATO);
        }

        @Override
        public void preCambiaStato(StatoOrdine.OrdineContestoStato context) {
            if (!context.actions().isBevandeInPreparazioneAllowed(context)) {
                throw new RuntimeException("for test purpose");
            }
        }

        @Override
        public void postCambiaStato(StatoOrdine.OrdineContestoStato context) {
            context.actions().bevandeInPreparazione(context);
        }
    },
    PRONTO {
        @Override
        public List<StatoOrdine> getProssimiStati() {
            return Arrays.asList(StatoOrdine.IN_CONSEGNA, StatoOrdine.CANCELLATO, StatoOrdine.NON_RITIRATO);
        }

        @Override
        public void postCambiaStato(StatoOrdine.OrdineContestoStato context) {
            context.actions().pronto(context);
            context.publish(Device.MONITOR_VASSOIO);
        }
    },
    IN_CONSEGNA {
        @Override
        public List<StatoOrdine> getProssimiStati() {
            return Arrays.asList(StatoOrdine.CONSEGNATO, StatoOrdine.CONSEGNATO_FORZATO);
        }

        @Override
        public void postCambiaStato(StatoOrdine.OrdineContestoStato context) {
            context.actions().inConsegna(context);
        }
    },
    //STATI Finali
    EVASO {
        @Override
        public void postCambiaStato(StatoOrdine.OrdineContestoStato context) {
            context.actions().evaso(context);
        }
    },
    EVASO_CONVENZIONE {
        @Override
        public List<StatoOrdine> getProssimiStati() {
            return Collections.singletonList(StatoOrdine.CHIUSO);
        }
    },
    CONSEGNATO {
        @Override
        public void postCambiaStato(StatoOrdine.OrdineContestoStato context) {
            context.publish(Device.POSTAZIONE_FORZATURA, Device.MONITOR_VASSOIO, Device.NUMERI, Device.NOTIFICA_CONSEGNATO);
        }
    },
    CONSEGNATO_FORZATO {
        @Override
        public void postCambiaStato(StatoOrdine.OrdineContestoStato context) {
            context.publish(Device.POSTAZIONE_FORZATURA, Device.MONITOR_VASSOIO, Device.NUMERI, Device.NOTIFICA_CONSEGNATO);
        }
    },
    NON_RITIRATO {
        @Override
        public List<StatoOrdine> getProssimiStati() {
            return Arrays.asList(StatoOrdine.RECUPERATO, StatoOrdine.IN_CONSEGNA, StatoOrdine.SCARTATO);
        }

        @Override
        public void postCambiaStato(StatoOrdine.OrdineContestoStato context) {
            context.actions().nonRitirato(context);
            context.publish(Device.MONITOR_VASSOIO);
        }
    },
    SCARTATO {
        @Override
        public void postCambiaStato(StatoOrdine.OrdineContestoStato context) {
            context.actions().scartato(context);
            context.publish(Device.MONITOR_VASSOIO);
        }
    };

    @AllArgsConstructor
    @NoArgsConstructor
    @Getter
    public static class OrdineContestoStato extends ContestoStato {
        private Ordine ordine;

        public OrdineContestoStato(Ordine o, BarLayout barLayout) {
            super.barLayout = barLayout;
            this.ordine = o;
        }

        public ActionStati actions() {
            return barLayout.getFlusso().getManager(ordine.getCanale());
        }

        public boolean isBeforeFineFasciaFreddi(Date date) {
            LocalTime limit = LocalTime.parse(barLayout.getOrarioAperturaCucina(), DateTimeFormatter.ofPattern("HH:mm"));
            LocalTime orderTime = LocalDateTime.ofInstant(Instant.ofEpochMilli(date.getTime()), ZoneId.systemDefault()).toLocalTime();
            return orderTime.isBefore(limit);
        }
    }

    public static List<StatoOrdine> stringsToStati(Collection<String> strings) {
        List<StatoOrdine> stati = new ArrayList<>();
        for (String s : strings) {
            stati.add(StatoOrdine.valueOf(s));
        }
        return stati;
    }

    public boolean cambiaStato(OrdineContestoStato context) {
        boolean statoModificato = false;

        Ordine ordine = context.getOrdine();

        StatoOrdine statoOrdine = ordine.getStato();

        if (ordine.getStato() != this && !statoOrdine.isStatoFinale()) {

            if (!statoOrdine.getProssimiStati().contains(this)) {
                throw new RuntimeException("for test purpose");
            }

            this.preCambiaStato(context);

            //TODO serve ancora?? se si spostarlo nel preCambiaStato
            if (context.getBarLayout().getFlusso() == FlussoServizio.BACK &&
                    (ordine.getCanale() == Canale.CHIOSCO || ordine.getCanale() == Canale.SERVITO && ordine.hasPiattiForPreparazioneAndStato(Arrays.asList(Piatto.Preparazione.CUCINA), Arrays.asList(StatoPiattoOrdinato.APERTO)))) {
                ordine.forzaAggiornamentoPiatti();
            }

            ordine.setStato(this);
            statoModificato = true;

            this.postCambiaStato(context);
        }

        return statoModificato;
    }

    public void preCambiaStato(StatoOrdine.OrdineContestoStato context) {
    }

    public void postCambiaStato(StatoOrdine.OrdineContestoStato context) {
    }

    @Override
    public List<StatoOrdine> getProssimiStati() {
        return Collections.emptyList();
    }

    @Override
    public boolean isStatoFinale() {
        return getProssimiStati().isEmpty();
    }

}