package net.lucamasini.testcase.shared.entity;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.google.common.base.Strings;
import lombok.*;
import net.lucamasini.testcase.shared.dto.Attesa;
import net.lucamasini.testcase.shared.dto.Canale;
import net.lucamasini.testcase.shared.dto.StatoOrdine;
import net.lucamasini.testcase.shared.dto.StatoPiattoOrdinato;
import org.jboss.errai.common.client.api.annotations.Portable;

import javax.persistence.*;
import javax.validation.constraints.NotNull;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.stream.Collectors;


@Builder
@Getter
@Setter
@AllArgsConstructor
@NoArgsConstructor
@Portable
@Entity
@EqualsAndHashCode(of = "barcode")
@Table(name = "ORDINE")
@NamedQueries({
        @NamedQuery(
                name = "OrdiniAppCheckedInChiusi",
                query = "select o, l from Ordine o join BarLayout l on o.bar = l.bar where o.canale = 'APP' and o.stato = 'CHIUSO' " +
                        "and o.dataInserimento > :today and o.dataCheckin is not null and :now > o.inizioFascia"),
        @NamedQuery(
                name = "OrdiniAppCheckedInPronti",
                query = "select o, l from Ordine o join BarLayout l on o.bar = l.bar where o.canale = 'APP' and o.stato = 'PRONTO' " +
                        "and o.dataInserimento > :today and o.dataCheckin is not null"),
        @NamedQuery(
                name = "OrdiniAppCheckedInNonRitirati",
                query = "select o, l from Ordine o join BarLayout l on o.bar = l.bar where o.canale = 'APP' and o.stato = 'NON_RITIRATO' " +
                        "and o.dataInserimento > :today and o.dataCheckin is not null"),
        @NamedQuery(
                name = "OrdiniAppScadutoCheckin",
                query = "select o, l from Ordine o join BarLayout l on o.bar = l.bar where o.canale = 'APP' and o.stato in ('PRONTO', 'CHIUSO') " +
                        "and o.dataInserimento > :today and o.dataCheckin is null and :now > o.fineFascia"),
        @NamedQuery(
                name = "OrdiniAperti",
                query = "select o, l from Ordine o join BarLayout l on o.bar = l.bar where o.stato = 'APERTO' " +
                        "and o.dataInserimento between :dateStart and :dateEnd"),
        @NamedQuery(
                name = "Ordine.findOrdiniAppByInizioFascia",
                query = "select o from Ordine o  where o.canale = 'APP' and o.inizioFascia = :fasciaValue and o.dataInserimento > :today" +
                        " and o.bar.id = :idBarValue and o.stato not in ('CONSEGNATO','CONSEGNATO_FORZATO')"),
        @NamedQuery(
                name = "Ordine.findOrdiniByCanaleAndStato",
                query = "select o from Ordine o  where o.canale in (:canali)  and o.stato = :statoValue and o.dataInserimento > :today" +
                        " and o.bar.id = :idBarValue"),
        @NamedQuery(
                name = "Ordine.findOrdiniScartabili",
                query = "select distinct o from Ordine o left join fetch o.piattiOrdinati where o.canale = 'APP'  and o.stato in ('CANCELLATO','NON_RITIRATO') and o.dataInserimento > :today" +
                        " and o.bar.id = :idBarValue")

})
@ToString(of = {"barcode"})
public class Ordine extends OrdbarGenericEntity {

    @Id
    String barcode;

    @NotNull
    @Enumerated(EnumType.STRING)
    StatoOrdine stato;

    @ManyToOne
    Bar bar;

    @Column(name = "NUMEROCOMANDACANALE")
    String numeroComanda;

    @Column(name = "NUMEROCOMANDA")
    Long oldNumeroComanda;

    @Transient
    boolean prioritario;

    public void updatePrioritario() {
        if (hasPiattiOrdinati()) {
            prioritario = piattiOrdinati.stream().filter(p -> p.piatto != null && Piatto.Preparazione.A_VISTA != p.piatto.getPreparazione())
                    .anyMatch(p -> Arrays.asList(StatoPiattoOrdinato.INCARICO, StatoPiattoOrdinato.PRONTO).contains(p.getStato()));
        }
    }

    //  Usare queste annotazioni commentate per far salvare automaticamente l'oggetto nelle relazioni
//  N.B. farà una update per ognuno
//  @OneToMany(cascade = CascadeType.ALL)
//  @JoinColumn(name = "ORDINE_ID")
//  Jackson tramite le annotazioni @JsonManagedReference e @JsonBackReference fa il set in automatico. Nei test non usiamo jackson!
    //@ValidOrdine
    @OneToMany(cascade = CascadeType.ALL, mappedBy = "ordine")
    @OrderBy("dataInserimento")
    @com.fasterxml.jackson.annotation.JsonManagedReference
    List<PiattoOrdinato> piattiOrdinati = new ArrayList<>();

    //@ValidOrdine
    @ElementCollection(fetch = FetchType.EAGER)
    @CollectionTable(name = "ORDINE_MENU", joinColumns = @JoinColumn(name = "ORDINE_ID"))
    List<Menu> menus = new ArrayList<>();

    @NotNull
    @Enumerated(EnumType.STRING)
    Canale canale;

    @NotNull
    Long idCatalogoOriginale;

    String cartaFidaty;
    String convenzione;
    boolean asporto;

    @Temporal(TemporalType.TIMESTAMP)
    Date dataChiusura;

    @Temporal(TemporalType.TIMESTAMP)
    Date dataPresaInCarico;

    @Temporal(TemporalType.TIMESTAMP)
    Date dataCheckin;

    @Temporal(TemporalType.TIMESTAMP)
    Date dataPronto;

    @Column(name = "ID_MONITOR")
    Long idMonitor;

    @Temporal(TemporalType.TIMESTAMP)
    private Date inizioFascia;

    @Temporal(TemporalType.TIMESTAMP)
    private Date fineFascia;

    Long numeroPersone;

    private String badge;

    @Override
    public void postPrePersist() {
        if (getStato() == null) {
            this.setStato(StatoOrdine.APERTO);
        }
    }

    public boolean hasPiattiOrdinati() {
        return getPiattiOrdinati() != null && !getPiattiOrdinati().isEmpty();
    }

    public boolean hasMenus() {
        return getMenus() != null && !getMenus().isEmpty();
    }

    public boolean hasOnlyPiattiOrdinatiAVista() {
        if (hasPiattiOrdinati()) {
            for (PiattoOrdinato p : getPiattiOrdinati()) {
                if (Arrays.asList(Piatto.Preparazione.CUCINA, Piatto.Preparazione.VASSOIO).contains(p.getPiatto().getPreparazione())) {
                    return false;
                }
            }
        }
        return true;
    }

    public boolean isForItemprem(StatoOrdine statoTarget) {
        return (StatoOrdine.EVASO_CONVENZIONE.equals(statoTarget) && Canale.SERVITO.equals(getCanale()) && !Strings.isNullOrEmpty(getConvenzione()))
                || StatoOrdine.CHIUSO.equals(statoTarget) && ((Canale.CHIOSCO.equals(getCanale()) && StatoOrdine.SOSPESO.equals(getStato())) || (Canale.SERVITO.equals(getCanale())));
    }

    public boolean blockDelete() {
        if (Canale.APP.equals(getCanale())) {
            return false;
        }
        return Arrays.asList(new StatoOrdine[]{StatoOrdine.BEVANDE_IN_PREPARAZIONE, StatoOrdine.PIATTI_IN_PREPARAZIONE}).contains(getStato());
    }


    @Override
    public void allineaRelazioni() {
        if (hasPiattiOrdinati()) {
            getPiattiOrdinati().forEach(p -> {
                if (p.getOrdine() == null) {
                    p.setOrdine(this);
                }
            });
        }
    }

    public void forzaAggiornamentoOrdine() {
        onPreUpdate();
    }

    public void forzaAggiornamentoPiatti() {
        if (hasPiattiOrdinati()) {
            getPiattiOrdinati().forEach(OrdbarGenericEntity::onPreUpdate);
        }
    }

    public void allineaConOrdineRemoto(Ordine ordineAggiornato) {
        setConvenzione(ordineAggiornato.getConvenzione());
        setStato(ordineAggiornato.getStato());
        setNumeroPersone(ordineAggiornato.getNumeroPersone());
        setAsporto(ordineAggiornato.isAsporto());

        if (ordineAggiornato.hasPiattiOrdinati()) {
            for (PiattoOrdinato piattoOrdinato : ordineAggiornato.getPiattiOrdinati()) {
                if (piattoOrdinato.getId() == null) {
                    piattiOrdinati.add(piattoOrdinato);
                }
            }
        }
        if (ordineAggiornato.hasMenus()) {
            setMenus(ordineAggiornato.getMenus());
        }
    }

    public Attesa getAttesa() {
        if (getStato() == StatoOrdine.PIATTI_IN_PREPARAZIONE && getDataPresaInCarico() != null) {
            return calcolaAttesa(getDataPresaInCarico());
        }
        return getStato() != StatoOrdine.CHIUSO || getDataChiusura() == null ? null : calcolaAttesa(Canale.APP.equals(getCanale())&& getDataCheckin()!=null?getDataCheckin():getDataChiusura());
    }

    private Attesa calcolaAttesa(Date startDate) {
        Date now = new Date();
        long diffMillis = now.getTime() - startDate.getTime();
        long minutes = diffMillis / 1000 / 60;

        // https://www.wolframalpha.com/input/?i=5+%2F+(0.1+*+x*x+%2B+5)+from+0+to+20

        double indiceDiAttesa = 5 / (0.1 * minutes * minutes + 5);
        return new Attesa(indiceDiAttesa, minutes);
    }

    public void aggiungiPiattoOrdinato(PiattoOrdinato piattoOrdinato) {
        if (Strings.isNullOrEmpty(piattoOrdinato.getPiatto().getSottoPreparazione())) {
            piattoOrdinato.getPiatto().setSottoPreparazione("PIATTI");
        }
        if (piattiOrdinati == null) {
            piattiOrdinati = new ArrayList<>();
        }
        piattiOrdinati.add(piattoOrdinato);
        piattoOrdinato.setOrdine(this);
    }

    @JsonIgnore
    public boolean isPiattiCucinaDaFare() {

        return getPiattiOrdinati().stream()
                .anyMatch(po -> po.getPiatto().getPreparazione() == Piatto.Preparazione.CUCINA && po.getStato() != StatoPiattoOrdinato.CANCELLATO
                        && po.getStato() != StatoPiattoOrdinato.EVASO);

    }


    public boolean hasPiattiForPreparazioneAndStato(List<Piatto.Preparazione> preparazioni, List<StatoPiattoOrdinato> stati) {
        boolean result = false;
        if (hasPiattiOrdinati()) {
            result = getPiattiOrdinati().stream().anyMatch(p -> preparazioni.contains(p.getPiatto().getPreparazione()) && stati.contains(p.getStato()));
        }
        return result;

    }


    public boolean hasPiattiForPreparazione(Piatto.Preparazione... preparazione) {

        List<Piatto.Preparazione> preparazioni = Arrays.stream(preparazione).collect(Collectors.toList());
        boolean result = false;
        if (hasPiattiOrdinati()) {
            result = getPiattiOrdinati().stream().anyMatch(p -> preparazioni.contains(p.getPiatto().getPreparazione()));
        }
        return result;

    }


    public boolean hasOnlyPiattiForPreparazione(Piatto.Preparazione... preparazione) {

        List<Piatto.Preparazione> preparazioni = Arrays.stream(preparazione).collect(Collectors.toList());
        boolean result = false;
        if (hasPiattiOrdinati()) {
            result = getPiattiOrdinati().stream().allMatch(p -> preparazioni.contains(p.getPiatto().getPreparazione()));
        }
        return result;

    }

    @JsonIgnore
    public boolean isFreddo() {
        return !hasPiattiForPreparazione(Piatto.Preparazione.CUCINA) && asporto;
    }


    @JsonIgnore
    public boolean isInFasciaCheckin(long tolleranzaInizio, long tolleranzaFine) {
        LocalDateTime now = LocalDateTime.now();
        LocalDateTime inizio = Instant.ofEpochMilli(inizioFascia.getTime()).atZone(ZoneId.systemDefault()).toLocalDateTime();
        LocalDateTime fine = Instant.ofEpochMilli(fineFascia.getTime()).atZone(ZoneId.systemDefault()).toLocalDateTime();

        return !now.isBefore(inizio.minusMinutes(tolleranzaInizio)) && !now.isAfter(fine.plusMinutes(tolleranzaFine));
    }

    //qui solo per la serializzazione
    public void setAttesa(Attesa attesa) {

    }

    @JsonIgnore
    public boolean hasSoloAVistaDaFare() {
        return getPiattiOrdinati().stream()
                .allMatch(po -> po.isForPreparazione(Piatto.Preparazione.A_VISTA) ||
                        po.getStato() == StatoPiattoOrdinato.CANCELLATO);
    }

    @JsonIgnore
    public Long getNumeroComandaWithOutPrefix() {
        if (numeroComanda != null) {
            return Long.valueOf(numeroComanda.substring(1));
        }
        return null;
    }

    public boolean isEntroTolleranzaFine(BarLayout l) {
        Date time = getDataCheckin() != null ? getDataCheckin() : new Date();
        Date fine = new Date(getFineFascia().getTime() + (l.getTolleranzaFineCheckIn() * 60000L));
        return time.before(fine);
    }

    public boolean isOrdineScartabile(BarLayout l) {
        return StatoOrdine.CANCELLATO == getStato() || (StatoOrdine.NON_RITIRATO == getStato() && !isEntroTolleranzaFine(l));
    }
}
