package net.lucamasini.testcase.shared.entity;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.databind.annotation.JsonSerialize;
import com.google.common.base.Strings;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import lombok.*;
import net.lucamasini.testcase.shared.dto.StatoPiattoOrdinato;
import net.lucamasini.testcase.shared.serializers.OrdineSerializer;
import org.jboss.errai.common.client.api.annotations.Portable;

import javax.persistence.*;
import javax.validation.Valid;
import javax.validation.constraints.NotNull;
import java.util.*;
import java.util.stream.Collectors;

import static java.util.stream.Collectors.groupingBy;

@Getter
@Setter
@Builder
@ToString
@NoArgsConstructor
@AllArgsConstructor(access = AccessLevel.PROTECTED)
@Portable
@Entity
@Table(name = "PIATTO_ORDINATO")
@NamedQueries({
        @NamedQuery(
                name = "PiattoOrdinato.PiattiApertiEInCarico",
                query = "select p from PiattoOrdinato p inner join p.ordine " +
                        "where (p.stato = 'APERTO' or (p.stato = 'INCARICO' and p.idMonitor  = :idMonitor)) " +
                        "and p.ordine.bar.id = :idBar " +
                        "and p.ordine.stato in (:statiOrdine) " +
                        "and p.piatto.preparazione in (:preparazioni) " +
                        "and p.dataInserimento > :today " +
                        "and ((p.stato = 'APERTO' and p.piatto.sottoPreparazione in (:sottoPreparazioni)) or (p.stato = 'INCARICO' and p.idMonitor  = :idMonitor))"),

        @NamedQuery(
                name = "PiattoOrdinato.PiattiInCarico",
                query = "select p from PiattoOrdinato p inner join p.ordine " +
                        "where p.ordine.bar.id = :idBar " +
                        "and p.ordine.stato in (:statiOrdine) " +
                        "and p.piatto.preparazione in (:preparazioni) " +
                        "and p.dataInserimento > :today " +
                        "and p.stato = 'INCARICO' and p.idMonitor  = :idMonitor"),


        @NamedQuery(
                name = "PiattoOrdinato.PiattoApertoPerIdArticoloOrigniale",
                query = "select p from PiattoOrdinato p inner join p.ordine " +
                        "where p.piatto.idArticoloOriginale = :idArticoloOriginale " +
                        "and p.ordine.bar.id = :idBar " +
                        "and p.ordine.barcode = :idOrdine " +
                        "and p.dataInserimento > :today")

})
public class PiattoOrdinato extends OrdbarGenericEntity {

    @Id
    @GeneratedValue(generator = "PIATTO_ORDINATO_SQ")
    @SequenceGenerator(name = "PIATTO_ORDINATO_SQ", sequenceName = "PIATTO_ORDINATO_SQ", allocationSize = 1)
    Long id;

    @Enumerated(EnumType.STRING)
    @Builder.Default
    StatoPiattoOrdinato stato = StatoPiattoOrdinato.APERTO;

    @Valid
    @NotNull
    @Embedded
    Piatto piatto;

    @ManyToOne
    @JoinColumn(name = "ORDINE_ID", referencedColumnName = "BARCODE")
    @JsonSerialize(using = OrdineSerializer.class)
    @com.fasterxml.jackson.annotation.JsonBackReference
    Ordine ordine;

    @NotNull
    Long quantita;
    long quantitaEvasa;
    String nota;

    @OneToMany(cascade = CascadeType.ALL, mappedBy = "piattoOrdinato", orphanRemoval = true)
    @OrderBy("dataInserimento")
    @com.fasterxml.jackson.annotation.JsonManagedReference
    @Builder.Default
    List<Abbinamento> abbinamenti = new ArrayList<>();

    String uuidMenu;

    boolean asporto;

    @Temporal(TemporalType.TIMESTAMP)
    Date dataPresaInCarico;

    @Temporal(TemporalType.TIMESTAMP)
    Date dataPronto;

    @Column(name = "ID_MONITOR")
    Long idMonitor;


    @Transient
    Abbinamento abbinamento;

    public PiattoOrdinato(PiattoOrdinato from) {
        this(from.getId(), from.getStato(), from.getPiatto(), from.ordine, from.getQuantita(), from.getQuantitaEvasa(),
                from.getNota(), from.getAbbinamenti(), from.getUuidMenu(), from.isAsporto(), from.getDataPresaInCarico(), from.getDataPronto(),
                from.getIdMonitor(), null);
        this.setVersion(from.getVersion());
        this.setDataAggiornamento(from.getDataAggiornamento());
        this.setDataInserimento(from.getDataInserimento());
    }

    @Override
    protected void postPrePersist() {
        if (getStato() == null) {
            this.setStato(StatoPiattoOrdinato.APERTO);
        }
    }

    @Override
    protected void postPreUpdate() {
        if (stato == StatoPiattoOrdinato.INCARICO && dataPresaInCarico == null) {
            dataPresaInCarico = new Date();
        }
        if (stato == StatoPiattoOrdinato.PRONTO && dataPronto == null) {
            dataPronto = new Date();
        }
    }

    @Override
    public void allineaRelazioni() {
        if (hasAbbinamenti()) {
            for (Abbinamento a : getAbbinamenti()) {
                if (a.getPiattoOrdinato() == null) {
                    a.setPiattoOrdinato(this);
                }
            }
        }
    }

    @JsonIgnore
    public long getQuantitaDaEvadere() {
        long quantitaDaEvadere = quantita - quantitaEvasa;
        return quantitaDaEvadere < 0 ? 0 : quantitaDaEvadere;
    }

    private boolean hasQuantitaDaEvadere() {
        return getQuantita() > getQuantitaEvasa();
    }

    private boolean hasZeroQuantita() {
        return getQuantita() == 0;
    }


    public boolean hasAbbinamenti() {
        return abbinamenti != null && !abbinamenti.isEmpty();
    }

    public void evadiAbbinamenti() {
        if (hasAbbinamenti()) {
            for (Abbinamento a : getAbbinamenti()) {
                if (a.getStato() == null || (a.getStato() != null && !Abbinamento.Stato.CANCELLATO.equals(a.getStato()))) {
                    a.setStato(Abbinamento.Stato.EVASO);
                }
            }
        }
    }

    public void cancellaAbbinamenti() {
        if (hasAbbinamenti()) {
            for (Abbinamento a : getAbbinamenti()) {
                a.setStato(Abbinamento.Stato.CANCELLATO);
            }
        }
    }

    public void allineaConPiattoOrdinatoRemoto(PiattoOrdinato piattoOrdinatoAggiornato) {
        allineaAbbinamentiPiattoRemoto(piattoOrdinatoAggiornato);

        getPiatto().setPreparazione(piattoOrdinatoAggiornato.getPiatto().getPreparazione());
        setQuantita(piattoOrdinatoAggiornato.getQuantita());

        setQuantitaEvasa(hasZeroQuantita() ? 0 : getQuantitaEvasa());
        if (hasZeroQuantita()) {
            cancellaAbbinamenti();
        }
        setNota(piattoOrdinatoAggiornato.getNota());
    }

    private void allineaAbbinamentiPiattoRemoto(PiattoOrdinato piattoOrdinatoAggiornato) {
        Map<Long, Multimap<Abbinamento.Stato, Abbinamento>> abbinamentiGroupped = new HashMap<>();
        if (hasAbbinamenti()) {  // raggruppo per nome e stato quelli sul db
            for (Abbinamento abbinamento : getAbbinamenti()) {
                if (!abbinamentiGroupped.containsKey(abbinamento.getIdAbbinamentoOriginale())) {
                    Multimap<Abbinamento.Stato, Abbinamento> abbinamentiPerStato = HashMultimap.create();
                    abbinamentiPerStato.put(abbinamento.getStato(), abbinamento);
                    abbinamentiGroupped.put(abbinamento.getIdAbbinamentoOriginale(), abbinamentiPerStato);
                } else {
                    if (!abbinamentiGroupped.get(abbinamento.getIdAbbinamentoOriginale()).containsKey(abbinamento.getStato())) {
                        abbinamentiGroupped.get(abbinamento.getIdAbbinamentoOriginale()).put(abbinamento.getStato(), abbinamento);
                    } else {
                        abbinamentiGroupped.get(abbinamento.getIdAbbinamentoOriginale()).get(abbinamento.getStato()).add(abbinamento);
                    }
                }
            }
        }

        if (piattoOrdinatoAggiornato.hasAbbinamenti()) {
            //rimuovo dalla mappa quelli che ci sono in entrambe gli oggetti e quelli che rimangono sono stati cancellati
            for (Abbinamento abbinamento : piattoOrdinatoAggiornato.getAbbinamenti()) {
                if (abbinamento.getId() != null) {
                    for (Abbinamento.Stato stato : Arrays.asList(Abbinamento.Stato.EVASO, Abbinamento.Stato.PRONTO, Abbinamento.Stato.APERTO)) {
                        if (abbinamentiGroupped.get(abbinamento.getIdAbbinamentoOriginale()) != null) {
                            List<Abbinamento> abbinamentoList = new ArrayList<>(abbinamentiGroupped.get(abbinamento.getIdAbbinamentoOriginale()).get(stato));
                            if (!abbinamentoList.isEmpty()) {
                                abbinamentiGroupped.get(abbinamento.getIdAbbinamentoOriginale()).get(stato).remove(abbinamentoList.get(0));
                                break;
                            }
                        }
                    }
                } else {
                    getAbbinamenti().add(abbinamento);
                }
            }
        }

        for (Multimap<Abbinamento.Stato, Abbinamento> multimaps : abbinamentiGroupped.values()) {
            for (Abbinamento abbinamento : multimaps.values()) {
                getAbbinamenti().remove(abbinamento);
            }
        }
    }

    //qui solo per la serializzazione
    public void setRoundedDataInserimento(Date butta) {
    }

    public Date getRoundedDataInserimento() {
        if (getDataInserimento() != null) {
            Date dateCopy = new Date(getDataInserimento().getTime());
            dateCopy.setSeconds(getDataInserimento().getSeconds() <= 30 ? 0 : 30);
            return dateCopy;
        }
        return null;
    }

    public boolean allAbbinamentiPronti() {
        if (hasAbbinamenti()) {
            return abbinamenti.stream().noneMatch(a -> Abbinamento.Stato.APERTO == a.getStato());
        }
        return true;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        PiattoOrdinato that = (PiattoOrdinato) o;
        return Objects.equals(id, that.id);
    }

    @Override
    public int hashCode() {
        return Objects.hash(id);
    }

    @JsonIgnore
    public PiattoOrdinatoOuterKey getPiattoOrdinatoBackOuterKey() {
        if (getPiatto() != null && getOrdine() != null) {
            return new PiattoOrdinatoBackOuterKey(
                    getPiatto().getNome(),
                    Strings.isNullOrEmpty(getNota()) ? "NONE" : getNota(),
                    isAsporto() || ordine.isAsporto());
        }
        return null;
    }

    @JsonIgnore
    public PiattoOrdinatoFrontOuterKey getPiattoOrdinatoFrontOuterKey() {
        if (getPiatto() != null && getOrdine() != null) {
            return new PiattoOrdinatoFrontOuterKey(
                    getPiatto().getNome(),
                    Strings.isNullOrEmpty(getNota()) ? "NONE" : getNota(),
                    isAsporto() || ordine.isAsporto(),
                    this.getRoundedDataInserimento());
        }
        return null;
    }

    @JsonIgnore
    public PiattoOrdinatoTodoKey getPiattoOrdinatoTodoKey() {
        if (getPiatto() != null && getOrdine() != null) {
            return new PiattoOrdinatoTodoKey(
                    getPiatto().getNome(),
                    ordine.isPrioritario() ? ordine.getNumeroComanda() : null);
        }
        return null;
    }

    @JsonIgnore
    public PiattoOrdianatoInnerKey getPiattoOrdianatoInnerKey() {
        if (getOrdine() != null) {
            return new PiattoOrdianatoInnerKey(
                    ordine,
                    getAbbinamento() != null ? getAbbinamento().getNome() : "NONE");
        }
        return null;
    }

    @JsonIgnore
    public boolean isEvadibile() {
        return getStato() != StatoPiattoOrdinato.CANCELLATO;
    }

    @JsonIgnore
    public boolean isForPreparazione(Piatto.Preparazione... preparazione) {
        List<Piatto.Preparazione> preparazioni = Arrays.stream(preparazione).collect(Collectors.toList());
        return preparazioni.contains(getPiatto().getPreparazione());
    }

    public void scartaAbbinamenti() {
        getAbbinamenti().stream().filter(a -> Abbinamento.Stato.CANCELLATO != a.getStato()).forEach(a -> a.setStato(Abbinamento.Stato.SCARTATO));
    }

    @JsonIgnore
    public StatoPiattoOrdinato getProssimoStato() {
        return hasQuantitaDaEvadere() ? StatoPiattoOrdinato.APERTO : hasZeroQuantita() ? StatoPiattoOrdinato.CANCELLATO : allAbbinamentiPronti() ? StatoPiattoOrdinato.PRONTO : getStato();
    }

    private static final String NONE = "NONE";

    public static Map<Date, Map<String, Map<String, Map<Boolean, Map<Ordine, Map<String, List<PiattoOrdinato>>>>>>> raggruppaPiattiDellaLista(List<PiattoOrdinato> piattiOrdinatiConAbbinamentiEsplosi) {
        return piattiOrdinatiConAbbinamentiEsplosi.stream().collect(
                groupingBy(
                        PiattoOrdinato::getRoundedDataInserimento,
                        groupingBy(
                                p -> p.getPiatto().getNome(),
                                groupingBy(
                                        p -> Strings.isNullOrEmpty(p.getNota()) ? NONE : p.getNota(),
                                        groupingBy(p -> p.isAsporto() || p.getOrdine().isAsporto(),
                                                groupingBy(
                                                        PiattoOrdinato::getOrdine,
                                                        groupingBy(
                                                                p -> p.getAbbinamento() != null ? p.getAbbinamento().getNome() : NONE)))))));
    }


}
